/* Priority queues, implemented as heaps
 * C version
 *
 * 15-122 Principles of Imperative Computation, Fall 2012
 * Frank Pfenning
 */

#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include "contracts.h"
#include "xalloc.h"
#include "heaps.h"

struct heap_header {
  int limit;			/* limit = capacity+1 */
  int next;			/* 1 <= next && next <= limit */
  elem* data;			/* \length(data) == limit */
};
typedef struct heap_header* heap;

/* priority(H, i) returns priority of element i in H
 * it should not assume H is a heap */
int priority(struct heap_header* H, int i) {
  REQUIRES(H != NULL);
  REQUIRES(1 <= i && i < H->next);
  return elem_priority(H->data[i]);
}

bool is_heap(struct heap_header* H) {
  if (!(H != NULL)) return false;
  //@assert \length(H->data) == H->limit;
  if (!(1 <= H->next && H->next <= H->limit)) return false;
  for (int i = 2; i < H->next; i++)
    //@loop_invariant 2 <= i;
    if (!(priority(H, i/2) <= priority(H, i))) return false;
  return true;
}

bool pq_empty(heap H) {
  REQUIRES(is_heap(H));
  return H->next == 1;
}

bool pq_full(heap H) {
  REQUIRES(is_heap(H));
  return H->next == H->limit;
}

heap pq_new(int capacity) { /* !! C !! */
  REQUIRES(capacity > 0);
  heap H = xcalloc(1, sizeof(struct heap_header));      /* !! C !! */
  H->limit = capacity+1;
  H->next = 1;
  H->data = xcalloc(capacity+1, sizeof(elem));          /* !! C !! */
  ENSURES(is_heap(H));
  return H;
}

/*void swap(void** A, int i, int j) {
  void* tmp = A[i];
  A[i] = A[j];
  A[j] = tmp;
  }*/

void swap(elem* A, int i, int j) {
  elem tmp = A[i];
  A[i] = A[j];
  A[j] = tmp;
}

void pq_insert(heap H, elem e) {
  REQUIRES(is_heap(H));
  REQUIRES(!pq_full(H));
  H->data[H->next] = e;
  (H->next)++;
  /* sift up to restore invariant */
  int i = H->next - 1;
  while (i > 1 && priority(H,i) < priority(H,i/2))
    {
      swap(H->data, i, i/2);
      i = i/2;
    }
  ENSURES(is_heap(H));
  return;
}

void sift_down(heap H, int i) {
  REQUIRES(1 <= i && i < H->next);
  int n = H->next;
  int left = 2*i;
  int right = left+1;
  while (left < n)
    //@loop_invariant 1 <= i && i < n;
    //@loop_invariant left == 2*i && right == 2*i+1;
    //@loop_invariant is_heap_except_down(H, i);
    { if (priority(H,i) <= priority(H,left)
	  && (right >= n || priority(H,i) <= priority(H,right)))
	return;
      if (right >= n || priority(H,left) < priority(H,right)) {
	swap(H->data, i, left);
	i = left;
      } else {
	//@assert right < n && priority(H, right) <= priority(H,left);
	swap(H->data, i, right);
	i = right;
      }
      left = 2*i;
      right = left+1;
    }
  //@assert i < n && 2*i >= n;
  //@assert is_heap_except_down(H, i);
  ENSURES(is_heap(H));
  return;
}

elem pq_delmin(heap H) {
  REQUIRES(is_heap(H));
  REQUIRES(!pq_empty(H));
  int n = H->next;
  elem min = H->data[1];
  H->data[1] = H->data[n-1];
  H->next = n-1;
  H->data[n-1] = NULL;		/* delete reference to element */
  if (H->next > 1) sift_down(H, 1);
  ENSURES(is_heap(H));
  return min;
}

elem pq_min(heap H) {
  REQUIRES(is_heap(H));
  REQUIRES(!pq_empty(H));
  return H->data[1];
}

void pq_free(heap H) {
  REQUIRES(is_heap(H));
  REQUIRES(pq_empty(H));
  free(H->data);
  free(H);
  return;
}
